@@ -1,5 +1,5 @@
 /* OpenVAS
-* $Id: pluginscheduler.c 9983 2011-01-17 17:46:44Z jan $
+* $Id: pluginscheduler.c 21159 2014-12-10 23:44:27Z kroosec $
 * Description: Tells openvassd which plugin should be executed next.
 *
 * Authors: - Renaud Deraison <deraison@nessus.org> (Original pre-fork develoment)
@@ -28,61 +28,59 @@
 */
 
 #include <string.h> /* for strcmp() */
-#include <stdio.h>  /* for printf() */
+
+#include <openvas/base/nvti.h>  /* for nvti_t */
 
 #include <openvas/misc/nvt_categories.h>  /* for ACT_SCANNER */
-#include <openvas/misc/plugutils.h>  /* for plug_get_required_ports */
-#include <openvas/misc/system.h>     /* for emalloc */
-#include <openvas/misc/arglists.h>
+#include <openvas/misc/plugutils.h>  /* for plug_get_launch */
+
+#include <openvas/base/nvticache.h>     /* for nvticache_t */
 
 #include <glib.h>
 
-#define IN_SCHEDULER_CODE 1
-
 #include "pluginscheduler.h"
 #include "pluginload.h"
 #include "pluginlaunch.h"
-#include "preferences.h"
 #include "log.h"
 
-
-/** @TODO
- * The pluginscheduler uses a name cache for NVTs, depedencies and ports, as
- * they are referenced by strings (names or OIDs). To remove duplicate code, the
- * pluginscheduler was changed to share the name cache with the arglists.
- * But the cache of the arglist is much smaller. It should be evaluated whether
- * this influenced the performance (memory imprint should be smaller), and
- * whether the string references could not be replaced e.g. by pointers to NVTi
- * structs.
- */
-
 /** @TODO
  * This important module needs documentation and comments.
  */
 
 #define HASH_MAX 2713
 
-
-/*-----------------------------------------------------------------------------*/
-
-int
-plugin_get_running_state (struct scheduler_plugin *plugin)
+struct hash
 {
-  return plugin->running_state;
-}
+  char *name;
+  struct scheduler_plugin *plugin;
+  gchar **dependencies;
+  int num_deps;
+  struct hash **dependencies_ptr;
+  gchar **ports;
+  struct hash *next;
+};
 
-void
-plugin_set_running_state (plugins_scheduler_t shed,
-                          struct scheduler_plugin *plugin, int state)
+struct list
 {
-  if (plugin == NULL)
-    return;
+  struct scheduler_plugin *plugin;
+  struct list *next;
+  struct list *prev;
+};
 
-  plugin->running_state = state;
-}
-
-/*-----------------------------------------------------------------------------*/
+struct plist
+{
+  gchar *name;
+  int occurences;
+  struct plist *next;
+  struct plist *prev;
+};
 
+struct plugins_scheduler
+{
+  struct hash *hash;                /**< Hash list of the plugins.   */
+  struct list *list[ACT_LAST + 1];  /**< Linked list of the plugins. */
+  struct plist *plist;              /**< Ports currently in use.     */
+};
 
 static unsigned int
 mkhash (char *name)
@@ -90,9 +88,6 @@ mkhash (char *name)
   return g_str_hash (name) % HASH_MAX;
 }
 
-/*------------------------------------------------------------------------------*/
-
-
 
 /*---------------------------------------------------------------------------*
  *
@@ -103,7 +98,7 @@ mkhash (char *name)
 static struct hash *
 hash_init ()
 {
-  struct hash *h = emalloc (sizeof (*h) * HASH_MAX + 1);
+  struct hash *h = g_malloc0 (HASH_MAX * sizeof (*h));
 
   return h;
 }
@@ -111,35 +106,24 @@ hash_init ()
 static void
 hash_link_destroy (struct hash *h)
 {
-  int i;
   if (h == NULL)
     return;
 
   if (h->next != NULL)
     hash_link_destroy (h->next);
 
-  if (h->dependencies != NULL)
-    {
-      for (i = 0; h->dependencies[i] != NULL; i++)
-        {
-          cache_dec (h->dependencies[i]);
-        }
-      efree (&h->dependencies);
-    }
+  g_free (h->dependencies_ptr);
 
-  efree (&h->dependencies_ptr);
-  efree (&h->plugin);
+  g_strfreev (h->plugin->required_ports);
+  g_strfreev (h->plugin->required_udp_ports);
+  g_strfreev (h->plugin->required_keys);
+  g_strfreev (h->plugin->mandatory_keys);
+  g_strfreev (h->plugin->excluded_keys);
+  g_free (h->plugin);
 
-  if (h->ports != NULL)
-    {
-      for (i = 0; h->ports[i] != NULL; i++)
-        {
-          cache_dec (h->ports[i]);
-        }
-      efree (&h->ports);
-    }
-
-  efree (&h);
+  g_strfreev (h->dependencies);
+  g_strfreev (h->ports);
+  g_free (h);
 }
 
 static void
@@ -151,19 +135,18 @@ hash_destroy (struct hash *h)
     {
       hash_link_destroy (h[i].next);
     }
-  efree (&h);
+  g_free (h);
 }
 
 
-static int
+static void
 hash_add (struct hash *h, char *name, struct scheduler_plugin *plugin)
 {
-  struct hash *l = emalloc (sizeof (struct hash));
-  unsigned int idx = mkhash (name);
-  struct arglist *deps = plug_get_deps (plugin->arglist->value);
-  struct arglist *ports = plug_get_required_ports (plugin->arglist->value);
-  int num_deps = 0;
+  struct hash *l = g_malloc0 (sizeof (struct hash));
+  unsigned int idx = mkhash ((char *)nvticache_get_filename ((const char *)name));
+  nvti_t *nvti;
 
+  nvti = nvticache_get_by_oid_full (plugin->arglist->name);
   l->plugin = plugin;
   l->plugin->parent_hash = l;
   l->name = name;
@@ -171,54 +154,23 @@ hash_add (struct hash *h, char *name, struct scheduler_plugin *plugin)
   h[idx].next = l;
   l->dependencies_ptr = NULL;
 
-  if (deps == NULL)
-    l->dependencies = NULL;
+  if (nvti_required_ports (nvti) != NULL)
+    l->ports = g_strsplit (nvti_required_ports (nvti), ", ", 0);
   else
-    {
-      struct arglist *al = deps;
-      int i = 0;
-      while (al->next)
-        {
-          num_deps++;
-          al = al->next;
-        }
-      l->dependencies = emalloc ((num_deps + 1) * sizeof (char *));
-      al = deps;
-      while (al->next != NULL)
-        {
-          l->dependencies[i++] = cache_inc (al->name);
-          l->num_deps++;
-          al = al->next;
-        }
-    }
-
-  if (ports == NULL)
     l->ports = NULL;
-  else
+
+  if (nvti_dependencies (nvti) != NULL)
     {
-      struct arglist *al = ports;
-      int num_ports = 0;
-      int i = 0;
-      while (al->next != NULL)
-        {
-          num_ports++;
-          al = al->next;
-        }
-
-      l->ports = emalloc ((num_ports + 1) * sizeof (char *));
-      al = ports;
-      while (al->next != NULL)
-        {
-          l->ports[i++] = cache_inc (al->name);
-          al = al->next;
-        }
+      l->dependencies = g_strsplit (nvti_dependencies (nvti), ", ", 0);
+      for (l->num_deps = 0; l->dependencies[l->num_deps] != NULL; l->num_deps ++) ;
     }
-  return 0;
+  else
+    l->dependencies = NULL;
+
+  nvti_free (nvti);
 }
 
 
-
-
 static struct hash *
 _hash_get (struct hash *h, char *name)
 {
@@ -226,7 +178,7 @@ _hash_get (struct hash *h, char *name)
   struct hash *l = h[idx].next;
   while (l != NULL)
     {
-      if (strcmp (l->name, name) == 0)
+      if (strcmp (nvticache_get_filename ((const char *)l->name), name) == 0)
         return l;
       else
         l = l->next;
@@ -243,9 +195,6 @@ hash_get_deps_ptr (struct hash *h, char *name)
   if (l == NULL)
     return NULL;
 
-  if (l->dependencies_ptr == NULL)
-    return NULL;
-
   return l->dependencies_ptr;
 }
 
@@ -253,40 +202,43 @@ static void
 hash_fill_deps (struct hash *h, struct hash *l)
 {
   int i, j = 0;
-  if (l->num_deps != 0)
+
+  if (l->num_deps == 0)
+    return;
+
+  l->dependencies_ptr =
+    g_malloc0 ((1 + l->num_deps) * sizeof (struct hash *));
+  for (i = 0; l->dependencies[i]; i++)
     {
-      l->dependencies_ptr =
-        emalloc ((1 + l->num_deps) * sizeof (struct hash *));
-      for (i = 0; l->dependencies[i]; i++)
+      struct hash *d = _hash_get (h, l->dependencies[i]);
+      if (d != NULL)
+        l->dependencies_ptr[j++] = d;
+      else
         {
-          struct hash *d = _hash_get (h, l->dependencies[i]);
+          gchar *path = g_path_get_dirname (nvticache_get_filename ((const char *)l->plugin->arglist->name));
+          if (g_ascii_strcasecmp (path, ".") != 0)
+            {
+              gchar *dep_with_path =
+                g_build_filename (path, l->dependencies[i], NULL);
+              d = _hash_get (h, dep_with_path);
+              g_free (dep_with_path);
+            }
+          g_free (path);
           if (d != NULL)
-            l->dependencies_ptr[j++] = d;
+            {
+              l->dependencies_ptr[j++] = d;
+            }
           else
             {
-              gchar *path = g_path_get_dirname (l->plugin->arglist->name);
-              if (g_ascii_strcasecmp (path, ".") != 0)
-                {
-                  gchar *dep_with_path =
-                    g_build_filename (path, l->dependencies[i], NULL);
-                  d = _hash_get (h, dep_with_path);
-                  g_free (dep_with_path);
-                }
-              g_free (path);
-              if (d != NULL)
-                {
-                  l->dependencies_ptr[j++] = d;
-                }
-              else
-                {
-                  log_write
-                    ("scheduler: %s depends on %s which could not be found, thus this dependency is not considered for execution sequence\n",
-                     l->plugin->arglist->name, l->dependencies[i]);
-                }
+              log_write ("scheduler: %s depends on %s which could not be found,"
+                         " thus this dependency is not considered for execution"
+                         " sequence",
+                         nvticache_get_filename ((const char *)l->plugin->arglist->name),
+                         l->dependencies[i]);
             }
         }
-      l->dependencies_ptr[j] = NULL;
     }
+  l->dependencies_ptr[j] = NULL;
 }
 
 /*----------------------------------------------------------------------*/
@@ -331,8 +283,8 @@ scheduler_mark_running_ports (plugins_scheduler_t sched,
         pl->occurences++;
       else
         {
-          pl = emalloc (sizeof (struct plist));
-          strncpy (pl->name, ports[i], sizeof (pl->name) - 1);  /* Share cache_inc() ? */
+          pl = g_malloc0 (sizeof (struct plist));
+          pl->name = g_strdup (ports[i]);
           pl->occurences = 1;
           pl->next = sched->plist;
           if (sched->plist != NULL)
@@ -372,23 +324,22 @@ scheduler_rm_running_ports (plugins_scheduler_t sched,
               else
                 sched->plist = pl->next;
 
-              efree (&pl);
+              g_free (pl->name);
+              g_free (pl);
             }
         }
       else
-        printf ("Warning: scheduler_rm_running_ports failed ?! (%s)\n",
-                ports[i]);
+        log_write ("Warning: scheduler_rm_running_ports failed ?! (%s)\n",
+                   ports[i]);
     }
 }
 
 
 struct scheduler_plugin *
-plugin_next_unrun_dependencie (plugins_scheduler_t sched,
-                               struct hash **dependencies_ptr,
-                               int already_in_dependencie)
+plugin_next_unrun_dependency (plugins_scheduler_t sched,
+                               struct hash **dependencies_ptr)
 {
   int flag = 0;
-  int counter = 0;
   int i;
 
   if (dependencies_ptr == NULL)
@@ -396,48 +347,49 @@ plugin_next_unrun_dependencie (plugins_scheduler_t sched,
 
   for (i = 0; dependencies_ptr[i] != NULL; i++)
     {
-      struct scheduler_plugin *plugin = dependencies_ptr[i]->plugin;
-      if (plugin != NULL)
+      struct scheduler_plugin *plugin;
+
+      plugin = dependencies_ptr[i]->plugin;
+      if (plugin == NULL)
+        continue;
+
+      switch (plugin->running_state)
         {
-          int state = plugin_get_running_state (plugin);
-          switch (state)
-            {
-            case PLUGIN_STATUS_UNRUN:
-              {
-                struct hash **deps_ptr = dependencies_ptr[i]->dependencies_ptr;
-                struct scheduler_plugin *ret;
-                counter++;
-                if (deps_ptr == NULL)
-                  return plugin;
-                else
-                  {
-                    ret = plugin_next_unrun_dependencie (sched, deps_ptr, 1);
-                    if (ret == NULL)
-                      return plugin;
-                    else if (ret == PLUG_RUNNING)
-                      flag++;
-                    else
-                      return ret;
-                  }
-            case PLUGIN_STATUS_RUNNING:
-                flag++;
-                break;
-            case PLUGIN_STATUS_DONE:
-                scheduler_rm_running_ports (sched, plugin);
-                plugin_set_running_state (sched, plugin,
-                                          PLUGIN_STATUS_DONE_AND_CLEANED);
-                break;
-            case PLUGIN_STATUS_DONE_AND_CLEANED:
-                break;
-              }
-            }
+        case PLUGIN_STATUS_UNRUN:
+          {
+            struct hash **deps_ptr;
+            struct scheduler_plugin *ret;
+
+            deps_ptr = dependencies_ptr[i]->dependencies_ptr;
+            if (deps_ptr == NULL)
+              return plugin;
+
+            ret = plugin_next_unrun_dependency (sched, deps_ptr);
+            if (ret == NULL)
+              return plugin;
+
+            if (ret == PLUG_RUNNING)
+              flag = 1;
+            else
+              return ret;
+          }
+          break;
+        case PLUGIN_STATUS_RUNNING:
+          flag = 1;
+          break;
+        case PLUGIN_STATUS_DONE:
+          scheduler_rm_running_ports (sched, plugin);
+          plugin->running_state = PLUGIN_STATUS_DONE_AND_CLEANED;
+          break;
+        case PLUGIN_STATUS_DONE_AND_CLEANED:
+          break;
         }
     }
 
-  if (flag == 0)
+  if (!flag)
     return NULL;
-  else
-    return PLUG_RUNNING;
+
+  return PLUG_RUNNING;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -447,36 +399,37 @@ plugin_next_unrun_dependencie (plugins_scheduler_t sched,
  */
 static void
 enable_plugin_and_dependencies (plugins_scheduler_t shed,
-                                struct arglist *plugin, char *name, int silent)
+                                struct arglist *plugin, char *name,
+                                GHashTable *deps_table)
 {
   struct hash **deps_ptr;
-  int i;
   int status;
+  int i;
 
   if (plugin == NULL)
     return;
 
+  if (g_hash_table_lookup (deps_table, name))
+    return;
+  else
+    g_hash_table_insert (deps_table, g_strdup (name), name);
+
   deps_ptr = hash_get_deps_ptr (shed->hash, name);
 
   status = plug_get_launch (plugin);
   if (status == LAUNCH_DISABLED)
-    {
-      if (silent == 0)
-        plug_set_launch (plugin, LAUNCH_RUN);
-      else
-        plug_set_launch (plugin, LAUNCH_SILENT);
-    }
+    plug_set_launch (plugin, LAUNCH_RUN);
 
-  if (deps_ptr != NULL)
+  if (deps_ptr == NULL)
+    return;
+  for (i = 0; deps_ptr[i] != NULL; i++)
     {
-      for (i = 0; deps_ptr[i] != NULL; i++)
-        {
-          struct scheduler_plugin *p;
-          p = deps_ptr[i]->plugin;
-          if (p != NULL && p->arglist != NULL)
-            enable_plugin_and_dependencies (shed, p->arglist->value,
-                                            p->arglist->name, silent);
-        }
+      struct scheduler_plugin *p;
+      p = deps_ptr[i]->plugin;
+      if (p != NULL && p->arglist != NULL)
+        enable_plugin_and_dependencies (shed, p->arglist->value,
+                                        (char *)nvticache_get_filename ((const char *)p->arglist->name),
+                                        deps_table);
     }
 }
 
@@ -484,45 +437,68 @@ enable_plugin_and_dependencies (plugins_scheduler_t shed,
 
 plugins_scheduler_t
 plugins_scheduler_init (struct arglist *plugins, int autoload,
-                        int silent_dependencies, int only_network)
+                        int only_network)
 {
-  plugins_scheduler_t ret = emalloc (sizeof (*ret));
+  plugins_scheduler_t ret;
   struct arglist *arg;
   int i;
   struct hash *l;
 
-
   if (plugins == NULL)
     return NULL;
 
-
   /* Fill our lists */
+  ret = g_malloc0 (sizeof (*ret));
   ret->hash = hash_init ();
   arg = plugins;
   while (arg->next != NULL)
     {
       struct scheduler_plugin *scheduler_plugin;
       struct list *dup;
-      int category = plug_get_category (arg->value);
+      char *oid = arg->name;
+      nvti_t *nvti;
+      int category;
 
-      scheduler_plugin = emalloc (sizeof (struct scheduler_plugin));
+      if (!oid)
+        continue; // This would be a serious problem as it shouldn't be
+      nvti = nvticache_get_by_oid_full (oid);
+      category = nvti_category (nvti);
+      scheduler_plugin = g_malloc0 (sizeof (struct scheduler_plugin));
       scheduler_plugin->arglist = arg;
       scheduler_plugin->running_state = PLUGIN_STATUS_UNRUN;
-      scheduler_plugin->category = plug_get_category (arg->value);
-      scheduler_plugin->timeout = plug_get_timeout (arg->value);
+      scheduler_plugin->category = category;
+      scheduler_plugin->timeout = nvti_timeout (nvti);
 
-      scheduler_plugin->required_ports = plug_get_required_ports (arg->value);
-      scheduler_plugin->required_udp_ports =
-        plug_get_required_udp_ports (arg->value);
-      scheduler_plugin->required_keys = plug_get_required_keys (arg->value);
-      scheduler_plugin->mandatory_keys = plug_get_mandatory_keys (arg->value);
-      scheduler_plugin->excluded_keys = plug_get_excluded_keys (arg->value);
+      if (nvti_required_ports (nvti) != NULL)
+        scheduler_plugin->required_ports = g_strsplit (nvti_required_ports (nvti), ", ", 0);
+      else
+        scheduler_plugin->required_ports = NULL;
 
+      if (nvti_required_udp_ports (nvti) != NULL)
+        scheduler_plugin->required_udp_ports = g_strsplit (nvti_required_udp_ports (nvti), ", ", 0);
+      else
+        scheduler_plugin->required_udp_ports = NULL;
+
+      if (nvti_required_keys (nvti) != NULL)
+        scheduler_plugin->required_keys = g_strsplit (nvti_required_keys (nvti), ", ", 0);
+      else
+        scheduler_plugin->required_keys = NULL;
+
+      if (nvti_mandatory_keys (nvti) != NULL)
+        scheduler_plugin->mandatory_keys = g_strsplit (nvti_mandatory_keys (nvti), ", ", 0);
+      else
+        scheduler_plugin->mandatory_keys = NULL;
+
+      if (nvti_excluded_keys (nvti) != NULL)
+        scheduler_plugin->excluded_keys = g_strsplit (nvti_excluded_keys (nvti), ", ", 0);
+      else
+        scheduler_plugin->excluded_keys = NULL;
+
+      nvti_free (nvti);
 
       if (category > ACT_LAST)
         category = ACT_LAST;
-      dup = emalloc (sizeof (struct list));
-      dup->name = scheduler_plugin->arglist->name;
+      dup = g_malloc0 ( sizeof (struct list));
       dup->plugin = scheduler_plugin;
       dup->prev = NULL;
       dup->next = ret->list[category];
@@ -534,7 +510,6 @@ plugins_scheduler_init (struct arglist *plugins, int autoload,
       arg = arg->next;
     }
 
-
   for (i = 0; i < HASH_MAX; i++)
     {
       l = &ret->hash[i];
@@ -550,23 +525,27 @@ plugins_scheduler_init (struct arglist *plugins, int autoload,
       arg = plugins;
       while (arg->next != NULL)
         {
+          /* deps_table is used to prevent circular dependencies. */
+          GHashTable *deps_table;
+
+          deps_table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+                                              NULL);
           if (plug_get_launch (arg->value) != LAUNCH_DISABLED)
-            enable_plugin_and_dependencies (ret, arg->value, arg->name,
-                                            silent_dependencies);
+            enable_plugin_and_dependencies
+             (ret, arg->value, (char *) nvticache_get_filename (arg->name),
+              deps_table);
           arg = arg->next;
+          g_hash_table_destroy (deps_table);
         }
     }
 
-
   /* Now, remove the plugins that won't be launched */
   for (i = ACT_FIRST; i <= ACT_LAST; i++)
     {
       struct list *l = ret->list[i];
       while (l != NULL)
         {
-          if (plug_get_launch (l->plugin->arglist->value) == LAUNCH_DISABLED
-              && plug_get_category (l->plugin->arglist->value) != ACT_INIT
-              && plug_get_category (l->plugin->arglist->value) != ACT_SETTINGS)
+          if (plug_get_launch (l->plugin->arglist->value) == LAUNCH_DISABLED)
             {
               struct list *old = l->next;
 
@@ -575,11 +554,10 @@ plugins_scheduler_init (struct arglist *plugins, int autoload,
               else
                 ret->list[i] = l->next;
 
-
               if (l->next != NULL)
                 l->next->prev = l->prev;
 
-              efree (&l);
+              g_free (l);
               l = old;
               continue;
             }
@@ -625,11 +603,7 @@ plugins_scheduler_next (plugins_scheduler_t h)
 
       while (l != NULL)
         {
-          int state;
-
-          state = plugin_get_running_state (l->plugin);
-
-          switch (state)
+          switch (l->plugin->running_state)
             {
             case PLUGIN_STATUS_UNRUN:
               {
@@ -639,20 +613,19 @@ plugins_scheduler_next (plugins_scheduler_t h)
                 if (deps_ptr != NULL)
                   {
                     struct scheduler_plugin *p =
-                      plugin_next_unrun_dependencie (h, deps_ptr, 0);
+                      plugin_next_unrun_dependency (h, deps_ptr);
 
                     switch (GPOINTER_TO_SIZE (p))
                       {
                       case GPOINTER_TO_SIZE (NULL):
                         scheduler_mark_running_ports (h, l->plugin);
-                        plugin_set_running_state (h, l->plugin,
-                                                  PLUGIN_STATUS_RUNNING);
+                        l->plugin->running_state = PLUGIN_STATUS_RUNNING;
                         return l->plugin;
 
                         break;
                       case GPOINTER_TO_SIZE (PLUG_RUNNING):
                         {
-                          /* One of the dependencie is still running  -  we write down its category */
+                          /* One of the dependency is still running  -  we write down its category */
                           if (l->plugin->category < running_category)
                             running_category = l->plugin->category;
                           flag++;
@@ -660,10 +633,9 @@ plugins_scheduler_next (plugins_scheduler_t h)
                         break;
                       default:
                         {
-                          /* Launch a dependencie  - don't pay attention to the type */
+                          /* Launch a dependency  - don't pay attention to the type */
                           scheduler_mark_running_ports (h, p);
-                          plugin_set_running_state (h, p,
-                                                    PLUGIN_STATUS_RUNNING);
+                          p->running_state = PLUGIN_STATUS_RUNNING;
                           return p;
                         }
                       }
@@ -671,8 +643,7 @@ plugins_scheduler_next (plugins_scheduler_t h)
                 else            /* No dependencies */
                   {
                     scheduler_mark_running_ports (h, l->plugin);
-                    plugin_set_running_state (h, l->plugin,
-                                              PLUGIN_STATUS_RUNNING);
+                    l->plugin->running_state = PLUGIN_STATUS_RUNNING;
                     return l->plugin;
                   }
               }
@@ -687,8 +658,7 @@ plugins_scheduler_next (plugins_scheduler_t h)
 
             case PLUGIN_STATUS_DONE:
               scheduler_rm_running_ports (h, l->plugin);
-              plugin_set_running_state (h, l->plugin,
-                                        PLUGIN_STATUS_DONE_AND_CLEANED);
+              l->plugin->running_state = PLUGIN_STATUS_DONE_AND_CLEANED;
               /* no break - we remove it right away */
             case PLUGIN_STATUS_DONE_AND_CLEANED:
               {
@@ -702,7 +672,7 @@ plugins_scheduler_next (plugins_scheduler_t h)
                 if (l->next != NULL)
                   l->next->prev = l->prev;
 
-                efree (&l);
+                g_free (l);
                 l = old;
 
                 continue;
@@ -739,7 +709,7 @@ list_destroy (struct list *list)
   while (list != NULL)
     {
       struct list *next = list->next;
-      efree (&list);
+      g_free (list);
       list = next;
     }
 }
@@ -752,5 +722,5 @@ plugins_scheduler_free (plugins_scheduler_t sched)
   hash_destroy (sched->hash);
   for (i = ACT_FIRST; i < ACT_LAST; i++)
     list_destroy (sched->list[i]);
-  efree (&sched);
+  g_free (sched);
 }
